#include <iostream>
#include <sstream>
#include <array>
#include <future>

template<int n> struct Int
{
    static const int val = n;
    auto operator++(   )const{ return Int<n+1>(); }
    auto operator++(int)const{ return Int<n+1>(); }
};

template<typename... Ts> struct List{};

template<typename L, typename R> struct Pair{ L l; R r; };

template<typename ic, typename ig, typename T0, typename... T> auto Idx_impl(ic icv, ig igv, T0&&,    T&&... ts){ return Idx_impl(++icv, igv, std::forward<T>(ts)...); }
template<typename ic,              typename T0, typename... T> auto Idx_impl(ic,     ic,     T0&& t0, T&&... ts){ return std::forward<T0>(t0); }

template<typename I, typename... Ts> auto Index( I i, Ts&&... ts ){ return Idx_impl( Int<0>(), i, std::forward<Ts>(ts)... ); }
template<typename I, typename... Ts> auto Index( I i, List<Ts...> l ){ return Idx_impl( Int<0>(), i, Ts()... ); }

template<typename Sel, typename... Coll> struct Sum{ decltype( Idx_impl(Int<0>(), Sel(), Coll()... ) ) val; };

template<template<typename, typename...> class SumType, typename Sel, typename... Coll, typename... Fs>
auto match( SumType<Sel, Coll...> const& sum, Fs&&... fs )
{
    return Index(Sel(), std::forward<Fs>(fs)...)( sum );
}

template<typename X> auto unfix( X const& ex ){ return ex; }
template<typename X> auto   fix( X const& ex ){ return ex; }

template<typename A, typename Tr>
auto cata( A const& alg, Tr const& tree )
{
    return alg( map( [&](auto const& x){ return cata(alg, x); }, unfix( tree ) ) );
}

template<typename CA, typename Tr>
auto ana( CA const& coalg, Tr const& tree )
{
    return fix( map( [&](auto const& x){ return ana(coalg, x); }, coalg( tree ) ) );
}

//-----------------------------------------------
template<typename T>struct Value{ T val; };

template<typename L, typename R>
auto Ctx( L const& l, R const& r ){ return Pair<L, R>{l, r}; }

using Lit = Int<0>;
using Var = Int<1>;
using Add = Int<2>;
using Mul = Int<3>;
using La  = Int<4>;
using App = Int<5>;
using Map = Int<6>;
using Arr = Int<7>;
using Idx = Int<8>;

template<int i> using Sym = Int<9+i>;

template<typename Op, typename...> struct Expression;
template<typename Op, typename E> struct Expression<Op, E>
{
    Op op; E e;
};

template<typename Op, typename Left, typename Right> struct Expression<Op, Left, Right>
{
    Op op; Left left; Right right;
};

template<typename T, int n> struct Expression<Arr, T, Int<n>>
{
    Arr op;
    T data[n];
    Int<n> size;
    decltype(auto) operator[](size_t i)     { return data[i]; }
    decltype(auto) operator[](size_t i)const{ return data[i]; }
};

template<typename T>
auto val( T const& x ){ return Expression<Lit, T>{Lit(), x}; }

template<typename Op, typename L, typename R> auto expr( Op const& op, L const& l, R const& r ){ return Expression<Op, L, R>{op, l, r}; }

template<typename S, typename E> auto la( S const& sym, E const& e ){ return Expression<La, S, E>{La(), sym, e}; }
template<typename E1, typename E2> auto app( E1 const& e1, E2 const& e2 ){ return Expression<App, E1, E2>{App(), e1, e2}; }
template<typename E1, typename E2> auto idx( E1 const& e1, E2 const& e2 ){ return Expression<Idx, E1, E2>{Idx(), e1, e2}; }
template<typename E1, typename E2> auto emap( E1 const& e1, E2 const& e2 ){ return Expression<Map, E1, E2>{Map(), e1, e2}; }
template<int i> auto sym(){ return Expression<Var, Sym<i>>{ Var(), Sym<i>() }; }

template<typename E1, typename E2> auto operator+( E1 const& e1, E2 const& e2 ){ return Expression<Add, E1, E2>{Add(), e1, e2}; }
template<typename E1, typename E2> auto operator*( E1 const& e1, E2 const& e2 ){ return Expression<Mul, E1, E2>{Mul(), e1, e2}; }

template<typename O1, typename O2, typename... E1s, typename... E2s>
auto add( Expression<O1, E1s...> const& e1, Expression<O2, E2s...> const& e2 ){ return Expression<Add, Expression<O1, E1s...>, Expression<O2, E2s...>>{Add(), e1, e2}; }
template<typename O1, typename O2, typename... E1s, typename... E2s>
auto mul( Expression<O1, E1s...> const& e1, Expression<O2, E2s...> const& e2 ){ return Expression<Mul, Expression<O1, E1s...>, Expression<O2, E2s...>>{Mul(), e1, e2}; }

template<typename T, typename... Ts>
auto arr( T const& t, Ts const& ... ts ){ return Expression<Arr, T, Int<sizeof...(ts)+1>>{ Arr(), {t, ts...}, Int<sizeof...(ts)+1>() }; }

template<typename L, typename A, typename... Etc>
auto arr_map( L const& l, A a, Etc const& ... etc )
{
    using T = decltype( l( a(etc...)[0], etc... ) );
    Expression<Arr, T, decltype(a(etc...).size)> res;
    auto n = res.size;
    for( int i=0; i<n.val; ++i ){ res[i] = l( a(etc...)[i], etc... ); }
    return res;
}

template<typename T, int n>
std::ostream& operator<<( std::ostream& ofstrm, Expression<Arr, T, Int<n>> const& e )
{
    ofstrm << "{";
    for(size_t i=0; i<e.size.val-1; ++i){ ofstrm << e[i] << ", "; }
    ofstrm << e[e.size.val-1] << "}";
    return ofstrm;
}

template<typename L, typename A, typename... Etc>
auto arr_map2( L const& l, A a, Etc const& ... etc )
{
    using T = decltype( l( a(etc...)[0], etc... ) );
    Expression<Arr, T, decltype(a(etc...).size)> res;
    auto n = res.size;
    std::array<std::future<T>, decltype(res.size)::val> handles;
    for( size_t i=0; i<n.val; ++i )
    {
        handles[i] = std::async( std::launch::async, [=](size_t idx)
        {
            std::stringstream ss;
            ss << "Thread (id=";
            std::this_thread::get_id()._To_text(ss);
            auto result =  l( a(etc...)[idx], etc... );
            ss << ") calculated: " << result << std::endl;
            std::cout << ss.str();
            return result;
        }, i );
    }
    for( size_t i=0; i<n.val; ++i ){ res[i] = handles[i].get(); }
    return res;
}

template<typename F, typename Op, typename... Exs>
auto map( F&& f, Expression<Op, Exs...> const& s )
{
    return match( s, [ ](auto const& lit_){ return lit_; },
                     [ ](auto const& var_){ return var_; },
                     [&](auto const& ex){ return expr(ex.op, f(ex.left), f(ex.right)); },
                     [&](auto const& ex){ return expr(ex.op, f(ex.left), f(ex.right)); },
                     [&](auto const& ex){ return expr(ex.op, ex.left,    f(ex.right)); },
                     [&](auto const& ex){ return expr(ex.op, f(ex.left), f(ex.right)); },
                     [&](auto const& ex){ return expr(ex.op, f(ex.left), f(ex.right)); },
                     [ ](auto const& ex){ return ex; },
                     [&](auto const& ex){ return expr(ex.op, f(ex.left), f(ex.right)); }
                );
}

template<typename S, typename C>                 auto get_( Expression<Var, S> const& e, C const& c0                                            ){ return e; }
template<typename S, typename C, typename... Cs> auto get_( Expression<Var, S> const& e, C const& c0,                          Cs const& ... cs ){ return get_(e, cs...); }
template<typename S, typename Q, typename... Cs> auto get_( Expression<Var, S> const& e, Pair<Expression<Var, S>, Q> const& p, Cs const& ... cs ){ return p.r; }
template<typename S, typename... Cs>             auto get ( Expression<Var, S> const& e, Cs const& ... cs ){ return get_(e, cs...); }

int main()
{
    auto x = sym<0>();
    auto y = sym<1>();
    auto ex = app( app( la(x, la(y, y+x*x ) ), val(5) ), val(4) );

    auto A = arr(arr(1, 2), arr(5, 6));
    auto ex2 = emap( la(x, emap( la(y, val(100) * idx(x, val(0)) + y*y ), x ) ), A );

    auto alg = []( auto const& ex )
    {
        return match( ex, [](auto const& lit_){ return [=](auto...){ return lit_.e; }; },
                          [](auto const& var_){ return [=](auto... x){ return get(var_, x...); }; },
                          [](auto const& add_){ return [=](auto... x){ return add_.left(x...) + add_.right(x...); }; },
                          [](auto const& mul_){ return [=](auto... x){ return mul_.left(x...) * mul_.right(x...); }; },
                          [](auto const& lam_){ return [=](auto q, auto... qq){ return lam_.right( qq..., Ctx(lam_.left, q) ); }; },
                          [](auto const& app_){ return [=](auto... x){ return app_.left( app_.right(), x... ); }; },
                          [](auto const& map_){ return [=](auto... x){ return arr_map2( map_.left, map_.right, x... ); }; },
                          [](auto const& arr_){ return [=](auto... x){ return arr_; }; },
                          [](auto const& idx_){ return [=](auto... x){ return idx_.left(x...)[ idx_.right(x...) ]; }; }
            );
    };

    auto res = cata(alg, ex2)();
    std::cout << "Starting array: " << A << std::endl;
    std::cout << "Result   array: " << res << std::endl;
}
